theory NikAxSoundComplete

imports NikAxiomatic

begin

(*
  We gave the rules to derive Hoare triplets: \<turnstile> {P} instr {Q}.
  Is this definition sound and complete?
    - soundness: if a Hoare triplet is derivable, it is valid?
    - completeness: if a Hoare triplet is valid, is it derivable?
*)

(* Soundness : if a Hoare triplet is derivable, it is valid *)
theorem axiomatic_sound:"\<turnstile> {P} instr {Q} \<Longrightarrow> \<Turnstile> {P} instr {Q}"
proof(induction rule: hoare_sem.induct)
  case (While P cond body) (* this is the only difficult case *)
  { 
    fix s t
    have "(while cond do body done, s) \<leadsto> t  \<Longrightarrow>  P s  \<Longrightarrow>  P t \<and> \<not> evalbool cond t"
    proof (induction "while cond do body done" s t rule: big_step_induct)
      case WhileFalse thus ?case by blast
    next
      case WhileTrue thus ?case
        using While.IH unfolding hoare_valid_def by blast
    qed
  }
  thus ?case unfolding hoare_valid_def by blast
qed (unfold hoare_valid_def, auto) (* the other cases are solved by auto *)

(*
  Completeness : every valid Hoare triplet is derivable
  For this, we need to introduce the notion of weakest precondition.
  For a given instruction and postcondition, this is the precondition
  which makes the postcondition hold on any state that can be reached
  by executing the instruction.
*)
(* Weakest precondition for validating postcondition Q after executing instruction instr *)
definition weakestpre :: "instruction \<Rightarrow> assertion \<Rightarrow> assertion"
where "weakestpre instr Q = (\<lambda>s. \<forall>t. (instr,s) \<leadsto> t  \<longrightarrow>  Q t)"

(* Find the weakest precondition for each instruction *)
lemma wp_Nop[simp]: "weakestpre Nop Q = Q"
unfolding weakestpre_def by auto

lemma wp_Decl[simp]: "weakestpre (var x) Q = Q[with (N 0) for x]"
unfolding weakestpre_def
apply (rule ext) (* two functions are equal if they take the same value for all inputs *)
apply (auto)
done

lemma wp_Assign[simp]: "weakestpre (x <- a) Q = Q[with a for x]"
unfolding weakestpre_def
apply (rule ext)
apply (auto)
done

lemma wp_Seq[simp]: "weakestpre (c\<^sub>1;;c\<^sub>2) Q = weakestpre c\<^sub>1 (weakestpre c\<^sub>2 Q)"
unfolding weakestpre_def
by auto

lemma wp_If[simp]:
 "weakestpre (if cond then iT else iF fi) Q =
 (\<lambda>s. if evalbool cond s then weakestpre iT Q s else weakestpre iF Q s)"
unfolding weakestpre_def
by auto

(* Show that the wekeast precondition does not change when unfolding a while loop *)
lemma wp_While_unfold:
 "weakestpre (while cond do body done) Q s =
  weakestpre (if cond then body ;; while cond do body done else Nop fi) Q s"
unfolding weakestpre_def
using unfold_while by auto

lemma wp_While_True[simp]: "evalbool cond s \<Longrightarrow>
  weakestpre (while cond do body done) Q s = weakestpre (body;; while cond do body done) Q s"
by(simp add: wp_While_unfold)

lemma wp_While_False[simp]: "\<not> evalbool cond s \<Longrightarrow> weakestpre (while cond do body done) Q s = Q s"
by(simp add: wp_While_unfold)

(* Show that the weakest precondition is a valid precondition *)
lemma weakestpre_is_pre: "\<turnstile> {weakestpre instr Q} instr {Q}"
proof (induction instr arbitrary: Q)print_cases
  case VarDecl thus ?case using hoare_sem.VDecl wp_Decl by presburger
next
  case Alternative thus ?case by (auto intro: Conseq)
next
  case (While cond body)
  let ?w = "while cond do body done"
  show "\<turnstile> {weakestpre ?w Q} ?w {Q}"
  proof(rule While')
    show "\<turnstile> {\<lambda>s. weakestpre ?w Q s \<and> evalbool cond s} body {weakestpre ?w Q}"
    proof(rule strengthen_pre[OF _ While.IH])
      show "\<forall>s. weakestpre ?w Q s \<and> evalbool cond s \<longrightarrow> weakestpre body (weakestpre ?w Q) s" by auto
    qed
    show "\<forall>s. weakestpre ?w Q s \<and> \<not> evalbool cond s \<longrightarrow> Q s" by auto
  qed
qed auto

(* We can now show that the axiomatic semantics of Niklaus is complete with regard 
   to the validity of Hoare triplets (defined in terms of big step semantics) *)
theorem axiomatic_complete:
  assumes "\<Turnstile> {P} instr {Q}"
  shows   "\<turnstile> {P} instr {Q}"
proof (rule strengthen_pre)
  show "\<forall>s. P s \<longrightarrow> weakestpre instr Q s" using assms
    by (simp add: hoare_valid_def weakestpre_def)
  show "\<turnstile> {weakestpre instr Q} instr {Q}" by (rule weakestpre_is_pre)
qed

(* So the axiomatic semantics is sound and complete *)
corollary axiomatic_sound_complete: "\<turnstile> {P} instr {Q} \<longleftrightarrow> \<Turnstile> {P} instr {Q}"
using axiomatic_sound and axiomatic_complete by auto

end
