theory NikAxiomatic

imports NikBigStep

begin

(* Type for an assertion on a state of the execution machine *)
type_synonym assertion = "state \<Rightarrow> bool"

(* Definition of a valid Hoare pre/post assertion *)
definition hoare_valid:: "assertion \<Rightarrow> instruction \<Rightarrow> assertion \<Rightarrow> bool" ("\<Turnstile> {_} _ {_}" 50)
where
  "\<Turnstile> {P} c {Q} = (\<forall>s t. P s \<and> (c, s) \<leadsto> t \<longrightarrow> Q t)"

(* Substitution of an expression to a variable in a state *)
abbreviation state_subst:: "expression \<Rightarrow> varname \<Rightarrow> state \<Rightarrow> state"
where "state_subst e v s \<equiv> s(v:= evaluate e s)"

(* Substitution of an expression to a variable in an assertion *)
abbreviation assert_subst:: "assertion \<Rightarrow> expression \<Rightarrow> varname \<Rightarrow> assertion" ("_[with _ for _]" [1000,0,0] 999)
where
  "P[with e for x] \<equiv> P \<circ> (state_subst e x)"

(* Semantic rules for the Hoare semantics of Niklaus *)
inductive hoare_sem:: "assertion \<Rightarrow> instruction \<Rightarrow> assertion \<Rightarrow> bool" ("\<turnstile> {_} _ {_}" 50)
where
  Nop:   "\<turnstile> {P} Nop {P}"
| VDecl: "\<turnstile> {P[with (N 0) for v]} var v {P}"
| Assign:"\<turnstile> {P[with u for v]} v <- u {P}"
| Seq:   "\<lbrakk> \<turnstile> {P} i1 {Q} ; \<turnstile> {Q} i2 {R} \<rbrakk> \<Longrightarrow> \<turnstile> {P} i1;; i2 {R}"
| If:    "\<lbrakk> \<turnstile> {\<lambda>s. P s \<and> (evalbool c s)} iT {Q}; \<turnstile> {\<lambda>s. P s \<and> \<not> (evalbool c s)} iF {Q} \<rbrakk>
          \<Longrightarrow> \<turnstile> {P} if c then iT else iF fi {Q}"
| While: "\<turnstile> {\<lambda>s. P s \<and> evalbool c s} body {P} 
          \<Longrightarrow> \<turnstile> {P} while c do body done {\<lambda>s. P s \<and> \<not>(evalbool c s)}"
| Conseq:"\<lbrakk>\<forall>s. P' s \<longrightarrow> P s; \<turnstile> {P} instr {Q}; \<forall>s. Q s \<longrightarrow> Q' s\<rbrakk> \<Longrightarrow> \<turnstile> {P'} instr {Q'}"

(* For proof automation, add these rules to simp and intro rules *)
lemmas[simp, intro!] = Nop VDecl Assign Seq If

(* Generate elimination rules *)
inductive_cases NopAxElim[elim!]:"\<turnstile> {P} Nop {P}"
inductive_cases VDeclAxElim[elim!]:"\<turnstile> {P[with (N 0) for v]} var v {P}"
inductive_cases AssignAxElim[elim!]:"\<turnstile> {P[with u for v]} v <- u {P}"
inductive_cases SeqAxElim[elim!]:"\<turnstile> {P} i1;; i2 {R}"
inductive_cases IfAxElim[elim!]:"\<turnstile> {P} if c then iT else iF fi {Q}"
inductive_cases WhileAxElim[elim]:"\<turnstile> {P} while c do body done {\<lambda>s. P s \<and> \<not>(evalbool c s)}"
inductive_cases ConseqAxElim[elim]:"\<turnstile> {P'} instr {Q'}"

(* The following two lemmas are easier to use than Conseq *)
(* We can always use a stronger precondition *)
lemma strengthen_pre:
  "\<lbrakk> \<forall>s. P' s \<longrightarrow> P s;  \<turnstile> {P} instr {Q} \<rbrakk> \<Longrightarrow> \<turnstile> {P'} instr {Q}"
by (rule Conseq, auto)

(* We can always get a weaker postcondition *)
lemma weaken_post:
  "\<lbrakk> \<turnstile> {P} instr {Q};  \<forall>s. Q s \<longrightarrow> Q' s \<rbrakk> \<Longrightarrow>  \<turnstile> {P} instr {Q'}"
by (rule Conseq, auto)

(* Small example *)
lemma "\<turnstile> {\<lambda>s. True} ''x'' <- N 2 {\<lambda>s. s ''x'' = 2}"
apply (rule strengthen_pre[of "\<lambda>s. True" "(\<lambda>s. (s ''x'' = 2))[with N 2 for ''x'']"])
apply simp
apply (rule Assign)
done

thm Assign
thm strengthen_pre
thm strengthen_pre[OF _ Assign]
(* Easier to use assignment rule *)
lemma Assign': "\<forall>s. P s \<longrightarrow> (Q[with u for v]) s \<Longrightarrow> \<turnstile> {P} v <- u {Q}"
apply (rule strengthen_pre[OF _ Assign])
apply (simp)
done

(* Same example with the new Assignment rule *)
lemma "\<turnstile> {\<lambda>s. True} ''x'' <- N 2 {\<lambda>s. s ''x'' = 2}"
apply (rule Assign')
apply simp
done

(* Same thing for the While rule *)
lemma While':
assumes "\<turnstile> {\<lambda>s. P s \<and> evalbool c s} body {P}" and "\<forall>s. P s \<and> \<not> evalbool c s \<longrightarrow> Q s"
shows "\<turnstile> {P} while c do body done {Q}"
by(rule weaken_post[OF While[OF assms(1)] assms(2)])

thm weaken_post
thm While
end
