theory NikBigStep
(*
 *  Big step operative semantics of Niklaus
 *)
imports NikInstructions

begin
(*
 *  In big step semantics, we define a relation between a pair (instruction, state) 
 *  and the state reached after executing the instruction from the original state.
 *  (i, s) and s' are in relation if by executing instruction i in state s we reach
 *  state s'.
 *
 *  Defining a relation r amounts to defining a predicate that takes two arguments and returns:
 *  - True is the arguments are in relation
 *  - False is they are not
 *
 *  Here, we use an inductive definition, which tells which pairs of (instruction, state)
 *  and state are in relation. We cannot use a function (see theory NikInstructions) 
 *  because we cannot prove its termination.
 *  The infix notation (i, s) \<leadsto> s', which states that (i, s) and s' are in relation,
 *  can be read as "executing i in state s leads to state s'".
 *)
inductive big_step :: "(instruction \<times> state) \<Rightarrow> state \<Rightarrow> bool" (infix "\<leadsto>" 55)
where
    Nop: "(Nop, s) \<leadsto> s" 
  | VDecl: "(var v, s) \<leadsto> s(v:=0)"
  | Assign: "(v <- e, s) \<leadsto> s(v:=evaluate e s)"
  | Seq: "\<lbrakk>(i1, s) \<leadsto> s'; (i2, s') \<leadsto> s''\<rbrakk>
          \<Longrightarrow> (i1;; i2, s) \<leadsto> s''"
  | IfTrue: "\<lbrakk>(evalbool c s); (instrTrue, s) \<leadsto> s'\<rbrakk>
         \<Longrightarrow> (if c then instrTrue else instrFalse fi, s) \<leadsto> s'"
  | IfFalse: "\<lbrakk>(\<not>evalbool c s); (instrFalse, s) \<leadsto> s'\<rbrakk>
         \<Longrightarrow> (if c then instrTrue else instrFalse fi, s) \<leadsto> s'"
  | WhileFalse: "(\<not>evalbool c s) \<Longrightarrow> (while c do instr done, s) \<leadsto> s"
  | WhileTrue: "\<lbrakk>(evalbool c s); (instr, s) \<leadsto> s';
                 (while c do instr done, s') \<leadsto> s''\<rbrakk>
             \<Longrightarrow> (while c do instr done, s) \<leadsto> s''"

print_theorems

(* Somme proofs to check that this works properly *)

(* P[of t1 t2 ... tn] substitutes t1, t2, ... tn to the n first schematic variables of P *)
thm VDecl
thm VDecl[of "''x''" "s"]
thm VDecl[where v="''x''" and s="s"]

lemma "(var ''x'', s) \<leadsto> s(''x'' := 0)"
by (rule VDecl[of "''x''" "s"])

(* The same lemma, but we let the unifier find substitutions *)
lemma "(var ''x'', s) \<leadsto> s(''x'' := 0)"
by (rule VDecl)

(* We can even let Isabelle find the rule by itself... *)
lemma "(var ''x'', s) \<leadsto> s(''x'' := 0)" ..

thm Assign
print_statement Assign
lemma "(''x'' <- N 3, s) \<leadsto> s(''x'' := 3)"
(* Here, we cannot apply the Assign rule directly, because this rule expects
   a term in the form "evaluate e pre". We introduce a subgoal of the right form: *)
apply (subgoal_tac "(''x'' <- N 3, s) \<leadsto> s(''x'' := evaluate (N 3) s)")
(* By simplification, we discharge the first goal *)
apply (simp)
(* We can now apply the Assign rule *)
apply (rule Assign)
done

(* Same lemma, with an Isar structured proof *)
lemma "(''x'' <- N 3, s) \<leadsto> s(''x'' := 3)"
proof -
  from evaluate.simps(1)[symmetric] have "3 = evaluate (N 3) s"  .
  then show ?thesis using Assign[where v="''x''" and e="(N 3)"]  by simp
qed

(* The same lemma again, with a less detailed Isar structured proof *)
lemma "(''x'' <- N 3, s) \<leadsto> s(''x'' := 3)"
proof -
  have "3 = evaluate (N 3) s"  by simp
  then show ?thesis using Assign[where e="(N 3)"] by simp
qed

(* In a schematic goal, schematic variables will be determined during the proof.
   Here, we will find the final state ?s'
 *)
schematic_goal ex1:  (* use schematic_lemma in versions of Isabelle previous to 2016 *)
  "(''x'' <- N 3 ;; ''y'' <- V ''x'', s) \<leadsto> ?s'"
(* Applying the Seq rule goes backward to the premisses of the lemma *)
apply (rule Seq)
(* The Assign rule discharges the first premisse *)
apply (rule Assign)
(* simp simplifies evaluate (N 3) s *)
apply (simp)
(* The Assign rule discharges the second premisse *)
apply (rule Assign)
done
(* ?s' was determined during this proof, we can get its value in the simplified form of the lemma *)
thm ex1
thm ex1[simplified]

(* Computing the GCD of 2 and 3 by successive subtractions *)
schematic_goal gcd2_3:
"   (''a''<- N 2;; ''b''<- N 3;;
    while V ''a'' .\<noteq>. V ''b'' do
      if V ''a'' .<. V ''b'' then
        ''b''<- V ''b'' .-.V ''a''
      else
        ''a''<- V ''a'' .-. V ''b''
      fi
    done,
    \<lambda>x.0) \<leadsto> ?s
"
apply(rule Seq)
apply(rule Assign)
apply simp
apply(rule Seq)
apply(rule Assign)
apply simp
apply(rule WhileTrue)
apply simp
apply(rule IfTrue)
apply simp
apply(rule Assign)
apply simp
apply(rule WhileTrue)
apply simp
apply(rule IfFalse)
apply simp
apply(rule Assign)
apply simp
apply(rule WhileFalse)
apply simp
done
thm gcd2_3

(* To avoid to much copy-pasting, we define an abbreviation *)
abbreviation "GCD_2_3 \<equiv> 
''a'' <- N 2 ;; ''b'' <- N 3 ;;
    while V ''a'' .\<noteq>. V ''b'' do
      if V ''a'' .<. V ''b'' then
        ''b'' <- V ''b'' .-. V ''a''
      else
        ''a'' <- V ''a'' .-. V ''b''
      fi
    done
"

(* Generate code for predicate big_step *)
code_pred big_step .

(* Compute the result of a program using this generated code *)
values "{res. (GCD_2_3, \<lambda>x.0) \<leadsto> res }"

(* Use map to display the value of a and b in the environment of the result *)
values "{map res [''a'', ''b'']|res. (GCD_2_3, \<lambda>x.0) \<leadsto> res }"

(* Equivalence of instructions *)
abbreviation equiv_instr :: "instruction \<Rightarrow> instruction \<Rightarrow> bool"
where "equiv_instr i1 i2 \<equiv> (\<forall> pre post. (i1, pre) \<leadsto> post = (i2, pre) \<leadsto> post)"

(* Proof automation *)
thm big_step.intros
declare big_step.intros [intro]
thm big_step.induct (* Standard induction rule *)
(* Allow splitting of arguments in lemmas *)
lemmas big_step_induct = big_step.induct[split_format(complete)]

(* Rule inversion *)
inductive_cases NopElim[elim!]: "(Nop, s) \<leadsto> s'"
thm NopElim
inductive_cases VDeclElim[elim!]: "(var x, s) \<leadsto> s'"
thm VDeclElim
inductive_cases AssignElim[elim!]: "(x <- e, s) \<leadsto> s'"
thm AssignElim
inductive_cases SeqElim[elim!]: "(i1;; i2, s) \<leadsto> s'"
thm SeqElim
inductive_cases AltElim[elim!]: "(if c then i1 else i2 fi, s) \<leadsto> s'"
thm AltElim
inductive_cases WhileElim[elim]: "(while c do body done, s) \<leadsto> s'"
thm WhileElim


lemma "(if c then Nop else Nop, pre) \<leadsto> post \<Longrightarrow> pre = post"
by auto

lemma "equiv_instr (if c then i else i fi) i"
by auto

lemma unfold_while: "equiv_instr (while cond do body done)
                   (if cond then body;; while cond do body done else Nop fi)"
by auto

(* Prove that the big step semantics is deterministic *)
lemma Nik_deterministic: "\<lbrakk> (prog, s) \<leadsto> s';
                            (prog, s) \<leadsto> s''\<rbrakk>
                         \<Longrightarrow> s'' = s'"
by (induction arbitrary: s'' rule: big_step_induct, blast+)

(* More detailed proof: we give some hints about why it is true in the different cases *)
lemma Nik_deterministic2: "\<lbrakk> (prog, s) \<leadsto> s';
                             (prog, s) \<leadsto> s''
                           \<rbrakk> \<Longrightarrow> s'' = s'"
proof (induction arbitrary: s'' rule: big_step.induct[split_format(complete)])
  print_cases
  case Nop thus ?case by (rule NopElim)
next
  case VDecl thus ?case by (rule VDeclElim)
next
  case Assign thus ?case by (rule AssignElim)
next
  case Seq  print_facts
  from Seq.prems[THEN SeqElim, OF Seq.IH(1)]
   and Seq.prems[THEN SeqElim, OF Seq.IH(2)] show ?case by simp
next
  case IfTrue print_facts
  from IfTrue.prems[THEN AltElim, OF IfTrue.IH] show ?case by (simp add: IfTrue(1))
next
  case IfFalse print_facts
  from IfFalse.prems[THEN AltElim, OF IfFalse.IH] show ?case by (simp add: IfFalse(1,3,4))
next
  case WhileFalse print_facts
  show ?case using WhileFalse.hyps WhileFalse.prems by auto
next
  case WhileTrue print_facts
  show ?case using WhileTrue.IH WhileTrue.hyps(1) WhileTrue.prems by auto
qed

(* Very detailed proof, almost like on the blackboard. *)
lemma Nik_deterministic3: "\<lbrakk> (prog, s) \<leadsto> s';
                             (prog, s) \<leadsto> s''
                           \<rbrakk> \<Longrightarrow> s'' = s'"
proof (induction arbitrary: s'' rule: big_step.induct[split_format(complete)])
  print_cases
  case (Nop pre)
  from `(Nop, pre) \<leadsto> s''` show "s'' = pre" by (rule NopElim)
next
  case (VDecl v pre)
  from `(var v, pre) \<leadsto> s''` show "s'' = pre(v := 0)" by (rule VDeclElim)
next
  case (Assign v e pre)
  from `(v <- e, pre) \<leadsto> s''` show "s'' = pre(v := evaluate e pre)" by (rule AssignElim)
next
  case (Seq i1 pre p' i2 post)
  from `(i1 ;; i2, pre) \<leadsto> s''` have "\<exists>q. (i1, pre) \<leadsto> q \<and> (i2, q) \<leadsto> s''" by blast
  then obtain q where q1:"(i1, pre) \<leadsto> q" and q2:"(i2, q) \<leadsto> s''" by blast
  from q1 and Seq.IH(1) have "q = p'" by simp
  with q2 and Seq.IH(2) show "s'' = post" by simp
next
  case (IfTrue c pre i1 post i2) print_facts
  from `(if c then i1 else i2 fi, pre) \<leadsto> s''` and `evalbool c pre` have "(i1, pre) \<leadsto> s''" by blast
  with IfTrue.IH show "s'' = post" .
next
  case (IfFalse c pre i2 post i1)
  from `(if c then i1 else i2 fi, pre) \<leadsto> s''` and `\<not> evalbool c pre` have "(i2, pre) \<leadsto> s''" by blast
  with IfFalse.IH show "s'' = post" .
next
  case (WhileFalse c pre body) print_facts
  from `(while c do body done, pre) \<leadsto> s''` and `\<not> evalbool c pre` show "s'' = pre" by blast
next
  case (WhileTrue c pre body p' post) print_facts
  from `(while c do body done, pre) \<leadsto> s''` and `evalbool c pre`
    have "\<exists>q. (body, pre) \<leadsto> q \<and> (while c do body done, q) \<leadsto> s''" by blast
  then obtain q where q1:"(body, pre) \<leadsto> q" and q2:"(while c do body done, q) \<leadsto> s''" by blast
  from q1 and WhileTrue.IH(1) have "q = p'" by simp
  with q2 and WhileTrue.IH(2) show "s'' = post" by simp
qed

end
