theory NikDenotational

imports NikInstructions

begin
(* The denotation of an instruction is a relation between states. *)
type_synonym instr_den = "state rel"

(*
  A functor to define the semantics of while as a least fixed point.
  This functor takes the condition and the denotation of the body of the while loop
  and returns a function which associates to a denotation:
    - the identity denotation if the condition evaluates to false
    - the composition of the denotation of the body and the denotation given 
      as argument if the condition evaluates to true.
  This definition matches Nop (whose denotation is the identity) when the 
  condition of the loop is false, and the sequence of the previous executions
  of the body and a new execution of the body when the condition is true.
  The denotation of the while loop is the least fixed point of this functor, 
  which is reached in a finite number of iterations of the functor when the
  loop terminates. Greater fixed points are not suitable as denotations of the
  while loop because they could have unwanted modifications of the state.

  O is the composition of relations (like o is the composition of functions):
  (a, b) \<in> r \<Longrightarrow> (b, c) \<in> s \<Longrightarrow> (a, c) \<in> r O s
*)
definition while_func:: "(state \<Rightarrow> bool) \<Rightarrow> instr_den \<Rightarrow> (instr_den \<Rightarrow> instr_den)"
where
  "while_func cond body_den = (\<lambda>w_den. {(pre, post).
                                 if cond pre then (pre, post) \<in> body_den O w_den
                                             else post = pre
                              })"

(* Existence of the fixed point: proving that while_func is monotone *)
lemma while_mono: "mono (while_func cond_den body_den)"
(* unfolding while_func_def mono_def by auto *)
apply(unfold while_func_def)
apply(unfold mono_def)
apply(auto)
done

(* More detailed proof *)
lemma while_mono2: "mono (while_func cond_den body_den)"
proof
  fix x::instr_den and y::instr_den
  assume "x \<subseteq> y"
  thus "while_func cond_den body_den x \<subseteq> while_func cond_den body_den y"
       unfolding while_func_def by auto
qed

fun nik_den:: "instruction \<Rightarrow> instr_den" ("\<lbrakk> _ \<rbrakk>")
where
  Nop_den:    "\<lbrakk>Nop\<rbrakk> = Id"
| Var_den:    "\<lbrakk>var v\<rbrakk> = {(s, s'). s' = s(v:=0)}"
| Assign_den: "\<lbrakk>v<-val\<rbrakk> = {(s, s'). s' = s(v:=evaluate val s)}"
| Seq_den:    "\<lbrakk>i1;; i2\<rbrakk> = \<lbrakk>i1\<rbrakk> O \<lbrakk>i2\<rbrakk>"
| Alt_den:    "\<lbrakk>if c then iT else iF fi\<rbrakk> =
                {(s, s').
                 if evalbool c s then
                   (s, s') \<in> \<lbrakk>iT\<rbrakk>
                 else
                   (s, s') \<in> \<lbrakk>iF\<rbrakk>
                }"
| While_den:  "\<lbrakk>while c do body done\<rbrakk> = lfp (while_func (evalbool c) \<lbrakk>body\<rbrakk>)"

lemma "\<lbrakk>''x''<-N 2\<rbrakk> = {(s, s'). s'=s(''x'':=2)}"
by simp

(* Consider an infinite loop over a Nop *)
abbreviation inf_nop where "inf_nop \<equiv> (while B True do Nop done)"

(* Show that the while functor applied to the denotation of this loop is the identity *)
lemma w_true_nop_id: "while_func (evalbool (B True)) \<lbrakk>Nop\<rbrakk> = (\<lambda>w. w)"
(* unfolding while_func_def by simp *)
proof -
  have "while_func (evalbool (B True)) \<lbrakk>Nop\<rbrakk> = while_func (\<lambda>s. True) Id"
    unfolding while_func_def using Nop_den evalbool.simps(7)[of "True"] by simp
  also have "... = (\<lambda>w. w)" unfolding while_func_def by simp
  finally show ?thesis .
qed

(*
  Therefore, any denotation is a fixed point of the while functor for this loop.
  Choosing the least one amounts to choosing the empty denotation: this loop
  never terminates so it defines no relation between states.
*)
thm lfp_lowerbound
lemma empty_lfp: "lfp (\<lambda>w. w) = {}"
proof
  have "(\<lambda>w. w) {} \<subseteq> {}" ..
  thus "lfp (\<lambda>w. w) \<subseteq> {}" using lfp_lowerbound[of "\<lambda>w. w" "{}"] by simp
next
  show "{} \<subseteq> lfp (\<lambda>w. w)" by simp
qed

lemma "\<lbrakk>while B True do Nop done\<rbrakk> = {}"
(* using empty_lfp w_true_nop_id by simp *)
proof -
  from nik_den.simps(6) have
    "\<lbrakk>inf_nop\<rbrakk> = lfp (while_func (evalbool (B True)) \<lbrakk>Nop\<rbrakk>)" by simp
  also from w_true_nop_id have "... = lfp (\<lambda>w .w)" by simp
  also from empty_lfp have "... = {}" .
  finally show ?thesis .
qed

(* Consider a no-loop over a Nop *)
abbreviation nop_loop where "nop_loop \<equiv> (while B False do Nop done)"

(* Show that the condition of the loop is always false *)
lemma w_nop_loop: "while_func (evalbool (B False)) \<lbrakk>Nop\<rbrakk> = while_func (\<lambda>s. False) Id"
unfolding while_func_def by simp

(* Show that this loop does nothing *)
lemma w_nop_loop_ok: "while_func (evalbool (B False)) \<lbrakk>Nop\<rbrakk> = (\<lambda>w. {(pre, post). post=pre})"
proof -
  from w_nop_loop have "while_func (evalbool (B False)) \<lbrakk>Nop\<rbrakk> = while_func (\<lambda>s. False) Id" .
  also have "... = (\<lambda>w. {(pre, post). post=pre})"
    unfolding while_func_def by simp
  finally show ?thesis .
qed


lemma "\<lbrakk>while B False do Nop done\<rbrakk> = \<lbrakk>Nop\<rbrakk>"
(* unfolding nik_den.simps(6) lfp_def using w_nop_loop_ok by auto *)
apply (insert nik_den.simps(6))
apply simp
apply (insert w_nop_loop_ok)
apply simp
apply (unfold lfp_def)
apply auto
done

end
