theory NikExprSimp
(*
 * Simplification of Niklaus arithmetic expressions
 *)
imports NikExpressions

begin

(*
 * Simplifying an expression amounts to replacing all constant expressions by their value.
 *)
fun simplify:: "expression \<Rightarrow> expression"
where
    (* A binary operator can be simplified if both arguments simplify to numbers *)
    "simplify (a .+. b) = (
       let (u', v') = (simplify a, simplify b) in
       if u' = (N 0) then v'
       else if v' = (N 0) then u'
       else
       case (u', v') of
           (N va, N vb) \<Rightarrow> N (va + vb)  (* arguments are numbers \<Rightarrow> compute the result *)
         | (sa, sb) \<Rightarrow> sa .+. sb        (* else return the sum of the simplified arguments *)
    )"
  | "simplify (a .-. b) = (
       let (u', v') = (simplify a, simplify b) in
       if v' = (N 0) then u'
       else
       case (u', v') of
           (N va, N vb) \<Rightarrow> N (va - vb)
         | (sa, sb) \<Rightarrow> sa .-. sb
    )"
  | "simplify (a .*. b) = (
       let (u', v') = (simplify a, simplify b) in
       if u' = (N 0) \<or> v' = (N 0) then N 0
       else if u' = (N 1) then v'
       else if v' = (N 1) then u'
       else
       case (u', v') of
           (N va, N vb) \<Rightarrow> N (va * vb)
         | (sa, sb) \<Rightarrow> sa .*. sb
    )"
  | "simplify (a .\<div>. b) = (
       let (u', v') = (simplify a, simplify b) in
       if u' = (N 0) then N 0
       else if v' = (N 1) then u'
       else
       case (u', v') of
           (N va, N vb) \<Rightarrow> N (va div vb)
         | (sa, sb) \<Rightarrow> sa .\<div>. sb
    )"
  (* Other expressions (variables and constants) cannot be simplified and are left as is *)
  | "simplify expr = expr"

(* Some checks *)
value "simplify ((N 2) .*. (N 3))"
value "simplify (N 2 .*. N 1 .+. V ''x'')"
value "simplify (N 0 .*. (N 1 .+. V ''x''))"
value "simplify ((N 2 .-. N 2) .*. (N 1 .+. V ''x''))"

thm expression.split

(* Proof that simplify does not change the semantics of expressions *)
theorem "evaluate (simplify expr) env = evaluate expr env"
(* by (induction expr, auto split:expression.split) *)
proof (induction expr) print_cases
  case Value show ?case by simp
next
  case Variable show ?case by simp
next
  (* We find that the proof is automatic if split according 
     to the various kinds of expressions.
     Using induction, we would have to process each case for a,
     and in each such case, to process each case for b!
   *)
  case (Sum a b) thus ?case by (auto split: expression.split)
(* Complete the proof using the same method for the other operations *)
qed (auto split: expression.split)

(* A predicate telling whether an expression is optimal, i.e. does not
   contain any operator applied to constants. *)
fun optimal_expr :: "expression \<Rightarrow> bool"
where
    "optimal_expr (N a) = True"
  | "optimal_expr (V x) = True"
  | "optimal_expr (a .+. b) = (
      case (a, b) of 
        (N va, N vb) \<Rightarrow> False
      | (ea, eb) \<Rightarrow> (optimal_expr ea) \<and> (optimal_expr eb)
     )"
  | "optimal_expr (a .-. b) = (
      case (a, b) of 
        (N va, N vb) \<Rightarrow> False
      | (ea, eb) \<Rightarrow> (optimal_expr ea) \<and> (optimal_expr eb)
     )"
  | "optimal_expr (a .*. b) = (
      case (a, b) of 
        (N va, N vb) \<Rightarrow> False
      | (ea, eb) \<Rightarrow> (optimal_expr ea) \<and> (optimal_expr eb)
     )"
  | "optimal_expr (a .\<div>. b) = (
      case (a, b) of 
        (N va, N vb) \<Rightarrow> False
      | (ea, eb) \<Rightarrow> (optimal_expr ea) \<and> (optimal_expr eb)
     )"

(* Check that optimal_expr behaves as expected *)
value "optimal_expr (N 3)"
value "optimal_expr (N 3 .+. V ''x'')"
value "optimal_expr (N 3 .+. N 5)"
value "optimal_expr ((N 3 .+. V ''x'') .*. V ''y'')"
value "optimal_expr ((N 3 .+. N 2) .*. V ''y'')"

(* Prove that simplify yields optimal expressions *)
theorem "optimal_expr (simplify expr)"
proof (induction expr) print_cases
  case Value show ?case by simp
next
  case Variable show ?case by simp
next
  case (Sum a b) thus ?case
    (* Since simplification of an operator depends on the form
       of the simplication of its arguments, we use method case_tac
       to discriminate the case where simplify a = N va, and then 
       to discriminate the case where simplify b = N vb.
       Simplifying all the remaining subgoals is then enough to
       finish the proof.
     *)
    apply (case_tac "simplify a")
    apply (case_tac "simplify b")
    apply simp_all
    done
next
  case (Diff a b) thus ?case
    by (case_tac "simplify a", case_tac "simplify b", simp_all)
next
  case (Prod a b) thus ?case
    by (case_tac "simplify a", case_tac "simplify b", simp_all)
next
  case (Quot a b) thus ?case
    by (case_tac "simplify a", case_tac "simplify b", simp_all)
qed

end
