theory NikExpressions
(*
 *  Niklaus arithmetic expressions.
 *
 *  An expression maybe: 
 *  - a constant integer value,
 *  - a variable,
 *  - the sum, difference, product or quotient of two expressions
 *)
imports Main

begin
type_synonym val = int            (* type of the expression value *)
type_synonym varname = string     (* type of the name of a variable *)
(* For evaluating expressions with variables, we need an environment 
   that tells us the value of the variables. So an environment is
   a function from variable names to values. *)
type_synonym environment = "varname \<Rightarrow> val"

(* Abstract syntax of the expressions in Niklaus (aka meta-model) *)
datatype expression =
    Value val                  ("N")              (* so N 2 is the number 2 *)
  | Variable varname           ("V")              (* so V ''x'' is the variable ''x'' *)
  | Sum expression expression  (infixr ".+." 65)  (* we add dots to avoid confusion with *)
  | Diff expression expression (infixr ".-." 65)  (* the +, -, * and \<div> predefined operators *)
  | Prod expression expression (infixr ".*." 75)
  | Quot expression expression (infixr ".\<div>." 75)

(*
 *  Semantics of expressions
 *  We associate an integer value to an expression in an environment that
 *  gives the value of variables.
 *  This semantics is compositional: the semantics of an expression is
 *  given in terms of the semantics of its parts.
 *)
fun evaluate:: "expression \<Rightarrow> environment \<Rightarrow> val"
where
    "evaluate (N v) env = v"      (* the value of a number is this number *)
  | "evaluate (V n) env = env n"  (* the value of a varible is given by the environment *)
  (* Niklaus arithmetic operations are mapped to the usual operations on integers *)
  | "evaluate (a .+. b) env = evaluate a env + evaluate b env"
  | "evaluate (a .-. b) env = evaluate a env - evaluate b env"
  | "evaluate (a .*. b) env = evaluate a env * evaluate b env"
  | "evaluate (a .\<div>. b) env = evaluate a env div evaluate b env"

(* A few checks *)
value "evaluate (N 2 .+. N 3) (\<lambda>x. 0)"
value "evaluate (V ''x'' .+. N 3) ((\<lambda>u. 0)(''x'':=5))"

end

