theory NikSmallStep
(*
 *  Small steps operative semantics of Niklaus
 *)
imports NikInstructions

begin

(*
 *  In small steps semantics, we define a relation between pairs of instructions and states.
 *  (i, s) and (i', s') are in relation if by executing instruction i in state s we reach
 *  state s' and still have to execute instruction i'.
 *
 *  Defining a relation r amounts to defining a predicate that takes two arguments and returns:
 *  - True is the arguments are in relation
 *  - False is they are not
 *
 *  Here, we use an inductive definition, which tells which pairs of (instruction, state)
 *  are in relation. We cannot use a function (see theory NikOpNaive) because we cannot
 *  prove its termination.
 *  The infix notation (i, s) \<rightarrow> (i', s'), which states that (i, s) and (i', s') are in relation,
 *  can be read as "executing i in state s leads in one step to executing i' in state s'".
 *)
inductive small_step:: "(instruction \<times> state) \<Rightarrow> (instruction \<times> state) \<Rightarrow> bool" (infix "\<rightarrow>" 55)
where
  (* These elementary instructions are executed in one step, so they lead in one step to Nop *)
  VDecl: "(var v, s) \<rightarrow> (Nop, s(v:=0))"
| Assign: "(v <- val, s) \<rightarrow> (Nop, s(v:=evaluate val s))"
  (* Chaining a Nop with another instruction *)
| SeqNop: "(Nop;; i, s) \<rightarrow> (i, s)"
  (* This one is more complex because instruction i may not be executable in one step *)
| Seq: "(i, s) \<rightarrow> (z, s') \<Longrightarrow> (i;; j, s) \<rightarrow> (z;; j, s')"
  (* Two cases for if_then_else: when the condition is true and when the condition is false *)
| IfTrue: "evalbool c s \<Longrightarrow> (if c then i1 else i2 fi, s) \<rightarrow> (i1, s)"
| IfFalse: "\<not>evalbool c s \<Longrightarrow> (if c then i1 else i2 fi, s) \<rightarrow> (i2, s)"
  (* While: when the condition is true, unroll the loop: execute the body, then the loop again *)
| WhileTrue: "evalbool c s \<Longrightarrow> 
    (while c do body done, s) \<rightarrow> (body;; while c do body done, s)"
  (* While: when the condition is false, become a Nop *)
| WhileFalse: "\<not>evalbool c s \<Longrightarrow>
    (while c do body done, s) \<rightarrow> (Nop, s)"

(* Generate standard ML code for this predicate *)
code_pred small_step .

(* The following builds the set of all 'post' such as var ''x'' executed in one step leads to post *)
values "{post. (var ''x'', \<lambda>x.0) \<rightarrow> post}"
(* Now, we look at the value of variables ''x'' and ''y'' in post. Since post = (instr, (env, input, output)),
   snd post is (env, input, output), and fst (snd post) is env. We use "map" to apply env to the list [''x'', ''y'']
   to get the value of varaibles x and y. *)
values "{map (snd post) [''x'', ''y'']|post. (var ''x'', \<lambda>x.0) \<rightarrow> post}"
(* Same thing to observe the result of x := 2 *)
values "{map (snd post) [''x'', ''y'']|post. (''x'' <- N 2, \<lambda>x.0) \<rightarrow> post}"

(* Here, we see that we execute only one step, so variable y is not yet modified in the result *)
values "{post. (''x'' <- N 2 ;; ''y'' <- V ''x'', \<lambda>x.0) \<rightarrow> post}"
values "{map (snd post) [''x'', ''y'']|post. (''x'' <- N 2 ;; ''y'' <- V ''x'', \<lambda>x.0) \<rightarrow> post}"

(*
 * Chaining small steps
 * We want to compute the behavior of a program as (i, s) \<rightarrow> (i2, s2) \<rightarrow> (i3, s3) ... \<rightarrow> (i', s')
 * For this, we need to define the reflexive transitive closure of \<rightarrow> 
 *)
(* Reflexive transitive closure of a relation *)
inductive star :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool)"
for r where  (* relation r is fixed in this definition *)
    refl: "(star r) x x"   (* star r is reflexive *)
  | trans: "\<lbrakk>r x y ; (star r) y z\<rbrakk> \<Longrightarrow> (star r) x z" (* star r closes r by transitivity *)
(* Hide rules refl and trans because they have too generic names which may create conflicts *)
hide_fact (open) refl trans

thm star.induct
print_statement star.induct

(* Show that star r is transitive *)
lemma star_trans: "\<lbrakk>(star r) x y ; (star r) y z\<rbrakk> \<Longrightarrow> (star r) x z"
(* by (induction rule:star.induct, (simp add: star.trans)+) *)
proof (induction rule:star.induct)
(* (star r x z) if either x = z (star.refl), so we have (star r) x z, done *)
  case (refl x) thus ?case .
next
(* or r x y and (star r) y za (star.trans).
   With (star r) za z, we get (star r) y z (by IH)
   and with r x y and (star r) y z we get (star r) x z, done*)
  case (trans x y za) print_facts
    from `star r za z` and trans.IH have "star r y z" by simp
    with `r x y` show ?case by (rule star.trans)
qed

lemma "\<lbrakk>(star r) x y ; (star r) y z\<rbrakk> \<Longrightarrow> (star r) x z"
apply (induction rule:star.induct)
apply (assumption)
apply (erule star.trans)
apply simp
done

(* star.induct works only for r x y.
   Here, we build a version of the induction theorem for relations on
   tuplets, like: r (i, pre) (i', post), for it to work with small_step
 *)
thm star.induct
lemmas star_induct =
  star.induct[of "r:: 'a \<times> 'b \<Rightarrow> 'a \<times> 'b \<Rightarrow> bool", split_format(complete)]

(* For proof automation, make reflexivity of star a simplification and an introduction rule *)
thm star.refl
declare star.refl[simp,intro]

(* Make this obvious fact a simplification and introduction rule *)
lemma star_trans1[simp, intro]: "r x y \<Longrightarrow> (star r) x y"
(* by (simp add: star.trans) *)
proof -
  assume 1: "r x y"
  moreover have "(star r) y y" using star.refl .
  ultimately show ?thesis by (simp add: star.trans)
qed

(* Generate code for star *)
code_pred star .

(* We can now define small_steps as the reflexive transitive closure of small_step
   We note (i, s) \<rightarrow>* (i', s') the fact that we can go in several steps 
   from (i, s) to (i', s')  *)
abbreviation small_steps :: 
  "instruction \<times> state \<Rightarrow> instruction \<times> state \<Rightarrow> bool" (infix "\<rightarrow>*" 55)
where "small_steps pre post \<equiv> (star small_step) pre post"

(* schematic_lemma is like lemma but allows for schematic variables.
   Here, we use the proof of a schematic lemma to compute the result
   of executing "x := 2; y = x" in an empty environment. *)
schematic_goal "(''x'' <- N 2 ;; ''y'' <- V ''x'', \<lambda>x.0) \<rightarrow>* ?post"
apply (rule star.trans)  (* force the execution of one step *)
apply (rule Seq)         (* apply the Seq rule to go to the first instruction *)
apply (rule Assign)      (* apply the Assign rule to handle it *)
apply simp               (* simplify to get the computed value of x *)
apply (rule star.trans)  (* go to the next step *)
apply (rule SeqNop)      (* get rid of the Nop we got when stepping through the assignment *)
apply (rule star.trans)  (* go to the next step *)
apply (rule Assign)      (* apply the Assign rule to handle it *)
apply simp               (* simplify to get the computed value of y *)
apply (rule star.refl)   (* apply the reflexivity rule to finish and get ?post *)
done
(* so we can see that at the end, we have x and y = 2 *)

(* Here, we use the generated ML code to compute the set of 
   (i', s') (we look only at the value of x and y in each s')
   that can be reached through big steps. *)
values "{map (snd post) [''x'', ''y'']
  |post. (''x''<-(N 3);; ''y''<-(V ''x''), \<lambda>x.0) \<rightarrow>* post}"
(* We get them in order:
    [0, 0] is reached in 0 steps (because of reflexivity)
    [3, 0] is reached after executing x := 3 (with Nop as next instruction)
    [3, 0] is reached after executing the Nop to which x:=3 led
    [3, 3] is reached after executing y := x (leading to a Nop)
 *)

(* In order to automatically prove the next theorem, we need to 
   give some information to the automatic proof methods.
   Elimination rules allow these methods to transform facts
   into simpler facts. For instance P \<and> Q can be transformed
   into facts P and Q by eliminating \<and> (this is rule conjE).
   Since we defined \<rightarrow> as an inductive predicate, the inductive_cases
   command allows us to generate these rules automatically. We just 
   have to state that they should be used as elimination rules with 
   the [elim] attribute. [elim!] says that the rule can be used as
   much as possible. For the while loop, we use [elim] because
   "as much as possible" would be an infinite number of times!
   There is a "thm" statement after each generation so that you
   can easily have a look at the form of these rules.
*)
(* Rule inversion *)
inductive_cases NopE[elim!]: "(Nop, pre) \<rightarrow> cont"
thm NopE
inductive_cases VDeclE[elim!]: "(var x, s) \<rightarrow> (i, s')"
thm VDeclE
inductive_cases AssignE[elim!]: "(x <- e, s) \<rightarrow> (i, s')"
thm AssignE
inductive_cases SeqNopE[elim!]: "(Nop;; i2, s) \<rightarrow> (i, s')"
thm SeqNopE
inductive_cases SeqE[elim!]: "(i1;; i2, pre) \<rightarrow> cont"
thm SeqE
inductive_cases AltE[elim!]: "(if c then iT else iF fi, s) \<rightarrow> (i, s')"
thm AltE
inductive_cases WhileE[elim]: "(while c do body done, s) \<rightarrow> (i, s')"
thm WhileE

(* Prove that the small step semantics is deterministic:
   a given instruction executed in a given state always leads to the same state.
 *)
theorem small_step_deterministic:
"\<lbrakk>pre \<rightarrow> post; pre \<rightarrow> post'\<rbrakk> \<Longrightarrow> post' = post"
apply (induction arbitrary: post' rule: small_step.induct)
apply auto
done

end
